## $Id: browser.py 8256 2012-04-23 07:36:46Z henrik $
##
## Copyright (C) 2012 Uli Fouquet & Henrik Bettermann
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
##
from datetime import datetime
import httplib
import urllib
from xml.dom.minidom import parseString
import grok
from waeup.kofa.browser.layout import KofaPage, UtilityView
from waeup.kofa.accesscodes import create_accesscode
from waeup.uniben.students.interfaces import ICustomStudentOnlinePayment
from waeup.uniben.applicants.interfaces import ICustomApplicantOnlinePayment
from waeup.kofa.students.browser import write_log_message
from waeup.kofa.students.viewlets import RequestCallbackActionButton as RCABStudent
from waeup.kofa.applicants.viewlets import RequestCallbackActionButton as RCABApplicant
from waeup.uniben.students.utils import actions_after_student_payment
from waeup.uniben.applicants.utils import actions_after_applicant_payment
from waeup.uniben.interfaces import MessageFactory as _

#    Interswitch test account data:
#
#    Card Number: 6274807700000007
#    Expiry Date: July 2012
#    PIN: 0000

#    Card Number: 6278050000000007
#    Expiry Date: July 2012
#    PIN: 0000

PRODUCT_ID = '57'
SITE_NAME = 'xyz.waeup.org'
PROVIDER_ACCT = '2345'
PROVIDER_BANK_ID = '8'
PROVIDER_ITEM_NAME = 'Kofa Provider Fee'
INSTITUTION_ACCT = '1234'
INSTITUTION_BANK_ID = '9'
INSTITUTION_NAME = 'Sample University'
CURRENCY = '566'
PAY_ITEM_ID = '5700'
QUERY_URL =   'https://testwebpay.interswitchng.com/test_paydirect/services/TransactionQueryURL.aspx'
POST_ACTION = 'https://testwebpay.interswitchng.com/test_paydirect/webpay/pay.aspx'

HOST = 'testwebpay.interswitchng.com'
URL = '/test_paydirect/services/TransactionQueryWs.asmx'
httplib.HTTPConnection.debuglevel = 0


def collect_xml_data(payment, xmldict):
    xmldict['detail_ref'] = payment.p_id
    xmldict['provider_amt'] = 100 * payment.surcharge_1
    xmldict['provider_acct'] = PROVIDER_ACCT
    xmldict['provider_bank_id'] = PROVIDER_BANK_ID
    xmldict['provider_item_name'] = PROVIDER_ITEM_NAME
    xmldict['institution_amt'] = 100 * payment.amount_auth
    xmldict['institution_acct'] = INSTITUTION_ACCT
    xmldict['institution_bank_id'] = INSTITUTION_BANK_ID
    xmldict['institution_item_name'] = payment.p_category
    xmldict['institution_name'] = INSTITUTION_NAME
    # Interswitch amount is not part of the xml data
    xmltext = """<payment_item_detail>
<item_details detail_ref="%(detail_ref)s" college="%(institution_name)s" department="%(department)s" faculty="%(faculty)s">
<item_detail item_id="1" item_name="%(institution_item_name)s" item_amt="%(institution_amt)d" bank_id="%(institution_bank_id)s" acct_num="%(institution_acct)s" />
<item_detail item_id="2" item_name="%(provider_item_name)s" item_amt="%(provider_amt)d" bank_id="%(provider_bank_id)s" acct_num="%(provider_acct)s" />
</item_details>
</payment_item_detail>""" % xmldict
    xml_data = """<input type="hidden" name="xml_data" value='%s'  />""" % xmltext
    return xml_data

def SOAP_post(soap_action,xml):
    """Handles making the SOAP request.

    Further reading:
    http://testwebpay.interswitchng.com/test_paydirect/services/TransactionQueryWs.asmx?op=getTransactionData
    """
    h = httplib.HTTPConnection(HOST)
    headers={
        'Host':HOST,
        'Content-Type':'text/xml; charset=utf-8',
        'Content-Length':len(xml),
        'SOAPAction':'"%s"' % soap_action,
    }
    h.request('POST', URL, body=xml,headers=headers)
    r = h.getresponse()
    d = r.read()
    if r.status!=200:
        raise ValueError('Error connecting: %s, %s' % (r.status, r.reason))
    return d

def get_SOAP_response(product_id, transref):
    xml="""\
<?xml version="1.0" encoding="utf-8"?>
<soap:Envelope xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/">
  <soap:Body>
    <getTransactionData xmlns="http://tempuri.org/">
      <product_id>%s</product_id>
      <trans_ref>%s</trans_ref>
    </getTransactionData>
  </soap:Body>
</soap:Envelope>""" % (product_id, transref)
    result_xml=SOAP_post("http://tempuri.org/getTransactionData",xml)
    doc=parseString(result_xml)
    response=doc.getElementsByTagName('getTransactionDataResult')[0].firstChild.data
    return response

def query_interswitch(user, payment, view):
    ob_class = view.__implemented__.__name__
    sr = get_SOAP_response(PRODUCT_ID, payment.p_id)
    user.loggerInfo(ob_class, 'callback received: %s' % sr)
    wlist = sr.split(':')
    if len(wlist) != 7:
        view.flash(_('Invalid callback: ${a}',
            mapping = {'a': wlist}))
        user.loggerInfo(ob_class,'invalid callback: %s' % payment.p_id)
        return False
    payment.r_code = wlist[0]
    payment.r_desc = wlist[1]
    payment.r_amount_approved = float(wlist[2]) / 100
    payment.r_card_num = wlist[3]
    payment.r_pay_reference = wlist[5]
    if payment.r_code != '00':
        view.flash(_('Unsuccessful callback: ${a}', mapping = {'a': wlist[1]}))
        user.loggerInfo(ob_class,'unsuccessful callback: %s' % payment.p_id)
        payment.p_state = 'failed'
        return False
    total_amount_auth = (
        payment.amount_auth
        + payment.surcharge_1
        + payment.surcharge_2)
    if payment.r_amount_approved != total_amount_auth:
        view.flash(_('Wrong amount'))
        user.loggerInfo(ob_class,'successful callback but wrong amount: %s'
            % payment.p_id)
        payment.p_state = 'failed'
        return False
    if wlist[4] != payment.p_id:
        view.flash(_('Wrong transaction id'))
        user.loggerInfo(ob_class,'successful callback but wrong transaction id: %s'
            % payment.p_id)
        payment.p_state = 'failed'
        return False
    user.loggerInfo(ob_class,'successful callback: %s' % payment.p_id)
    payment.p_state = 'paid'
    payment.payment_date = datetime.now()
    return True


class InterswitchActionButtonStudent(RCABStudent):
    grok.order(2)
    grok.context(ICustomStudentOnlinePayment)
    icon = 'actionicon_pay.png'
    text = _('CollegePAY')
    target = 'goto_interswitch'

    @property
    def target_url(self):
        if self.context.p_state != 'unpaid':
            return ''
        return self.view.url(self.view.context, self.target)

class InterswitchActionButtonApplicant(RCABApplicant):
    grok.order(2)
    grok.context(ICustomApplicantOnlinePayment)
    icon = 'actionicon_pay.png'
    text = _('CollegePAY')
    target = 'goto_interswitch'

    @property
    def target_url(self):
        if self.context.p_state != 'unpaid':
            return ''
        return self.view.url(self.view.context, self.target)

# Deprecated
class InterswitchRequestCallbackActionButtonStudent(RCABStudent):
    grok.order(3)
    grok.context(ICustomStudentOnlinePayment)
    icon = 'actionicon_call.png'
    text = _('Request CollegePAY callback')

    def target_url(self):
        if self.context.p_state == 'paid':
            return ''
        site_redirect_url = self.view.url(self.view.context, 'isw_callback')
        args = {
            'transRef':self.context.p_id,
            'prodID':PRODUCT_ID,
            'redirectURL':site_redirect_url}
        return QUERY_URL + '?%s' % urllib.urlencode(args)

# Alternative preferred solution
class InterswitchRequestWebserviceActionButtonStudent(RCABStudent):
    grok.order(4)
    grok.context(ICustomStudentOnlinePayment)
    icon = 'actionicon_call.png'
    text = _('Request CollegePAY webservice')
    target = 'request_webservice'

class InterswitchRequestWebserviceActionButtonApplicant(RCABApplicant):
    grok.order(4)
    grok.context(ICustomApplicantOnlinePayment)
    icon = 'actionicon_call.png'
    text = _('Request CollegePAY webservice')
    target = 'request_webservice'


class InterswitchPageStudent(KofaPage):
    """ View which sends a POST request to the Interswitch
    CollegePAY payment gateway.
    """
    grok.context(ICustomStudentOnlinePayment)
    grok.name('goto_interswitch')
    grok.template('student_goto_interswitch')
    grok.require('waeup.payStudent')
    label = _('Submit data to CollegePAY (Interswitch Payment Gateway)')
    submit_button = _('Submit')
    action = POST_ACTION
    site_name = SITE_NAME
    currency = CURRENCY
    pay_item_id = PAY_ITEM_ID
    product_id = PRODUCT_ID

    def update(self):
        #if self.context.p_state != 'unpaid':
        if self.context.p_state == 'paid':
            self.flash(_("Payment ticket can't be re-send to CollegePAY."))
            self.redirect(self.url(self.context, '@@index'))
            return

        self.student = self.context.getStudent()
        certificate = getattr(self.student['studycourse'],'certificate',None)
        xmldict = {}
        if certificate is not None:
            xmldict['department'] = certificate.__parent__.__parent__.code
            xmldict['faculty'] = certificate.__parent__.__parent__.__parent__.code
        else:
            xmldict['department'] = None
            xmldict['faculty'] = None
        self.amount = (self.context.amount_auth + self.context.surcharge_1 +
            self.context.surcharge_2 + self.context.surcharge_3)
        self.amount_100 = 100 * self.amount
        self.local_date_time = str(self.context.creation_date)
        self.site_redirect_url = self.url(self.context, 'request_webservice')
        self.xml_data = collect_xml_data(self.context, xmldict)
        return

class InterswitchPageApplicant(InterswitchPageStudent):
    """ View which sends a POST request to the Interswitch
    CollegePAY payment gateway.
    """
    grok.context(ICustomApplicantOnlinePayment)
    grok.require('waeup.payApplicant')
    grok.template('applicant_goto_interswitch')

    def update(self):
        if self.context.p_state != 'unpaid':
            self.flash(_("Payment ticket can't be re-send to CollegePAY."))
            self.redirect(self.url(self.context, '@@index'))
            return
        self.applicant = self.context.__parent__
        xmldict = {}
        xmldict['department'] = None
        xmldict['faculty'] = None
        self.amount = (self.context.amount_auth + self.context.surcharge_1 +
            self.context.surcharge_2 + self.context.surcharge_3)
        self.amount_100 = 100 * self.amount
        self.local_date_time = str(self.context.creation_date)
        self.site_redirect_url = self.url(self.context, 'request_webservice')
        self.xml_data = collect_xml_data(self.context, xmldict)
        return

# Deprecated
class InterswitchPaymentCallbackPageStudent(UtilityView, grok.View):
    """ Callback view for the CollegePAY gateway
    """
    grok.context(ICustomStudentOnlinePayment)
    grok.name('isw_callback')
    grok.require('waeup.payStudent')

    # This view is not yet working for offline querying transactions
    # since the query string differs from the query string sent after
    # posting transactions. This Interswitch bug must be removed first.
    # Alternatively, we could use the webservice only and replace
    # the RequestCallbackActionButton by a RequestWebserviceActionButton

    def update(self):
        if self.context.p_state == 'paid':
            self.flash(_('This ticket has already been paid.'))
            return
        student = self.context.getStudent()
        query = self.request.form
        write_log_message(self,'callback received: %s' % query)

        self.context.r_card_num = query.get('cardNum', None)
        self.context.r_code = query.get('resp', None)
        self.context.r_pay_reference  = query.get('payRef', None)
        self.context.r_amount_approved = float(query.get('apprAmt', '0.0')) / 100
        self.context.r_desc = query.get('desc', None)

        if self.context.r_code != '00':
            self.flash(_('Unsuccessful callback: ${a}',
                mapping = {'a': query.get('desc', _('Incomplete query string.'))}))
            write_log_message(self,'unsuccessful callback: %s' % self.context.p_id)
            self.context.p_state = 'failed'
            return

        total_amount_auth = (
            self.context.amount_auth
            + self.context.surcharge_1
            + self.context.surcharge_2)

        if self.context.r_amount_approved != total_amount_auth:
            self.flash(_('Wrong amount'))
            write_log_message(
                self,'successful but wrong amount: %s' % self.context.p_id)
            self.context.p_state = 'failed'
            return

        try:
            validation_list = get_SOAP_response(
                PRODUCT_ID, self.context.p_id).split(':')
            # Validation does not make sense yet since the query string
            # formats are conflicting. We are only printing the validation
            # string, nothing else.
            print 'WARNING: Webservice validation is not yet implemented'
            print 'validation list: %s' % validation_list
        except:
            print 'Connection to webservice failed.'

        # Add webservice validation here

        write_log_message(self,'valid callback: %s' % self.context.p_id)
        self.context.p_state = 'paid'
        self.context.payment_date = datetime.now()
        actions_after_student_payment(student, self.context, self)
        return

    def render(self):
        self.redirect(self.url(self.context, '@@index'))
        return

# Alternative solution, replaces InterswitchPaymentCallbackPage
class InterswitchPaymentRequestWebservicePageStudent(UtilityView, grok.View):
    """ Request webservice view for the CollegePAY gateway
    """
    grok.context(ICustomStudentOnlinePayment)
    grok.name('request_webservice')
    grok.require('waeup.payStudent')

    def update(self):
        if self.context.p_state == 'paid':
            self.flash(_('This ticket has already been paid.'))
            return
        student = self.context.getStudent()
        if query_interswitch(student, self.context, self):
            actions_after_student_payment(student, self.context, self)
        return

    def render(self):
        self.redirect(self.url(self.context, '@@index'))
        return

class InterswitchPaymentRequestWebservicePageApplicant(UtilityView, grok.View):
    """ Request webservice view for the CollegePAY gateway
    """
    grok.context(ICustomApplicantOnlinePayment)
    grok.name('request_webservice')
    grok.require('waeup.payApplicant')

    def update(self):
        if self.context.p_state == 'paid':
            self.flash(_('This ticket has already been paid.'))
            return
        applicant = self.context.__parent__
        if query_interswitch(applicant, self.context, self):
            actions_after_applicant_payment(student, self.context, self)
        return

    def render(self):
        self.redirect(self.url(self.context, '@@index'))
        return